
// ***********************************************************************
//  Module:    rust_tp                                                   *
//                                                                       *
//  Purpose:   Example transaction program using CPI-C API to respond    *
//             to pings received.                                        *
//                                                                       *
// To compile this program use the following cargo build commands:       *
//                                                                       *
// - on Linux                                                            *
//   export LD_LIBRARY_PATH=/opt/ibm/sna/lib64:/usr/lib64                *
//   export LD_RUN_PATH=/opt/ibm/sna/lib64:/usr/lib64                    *
//   RUSTFLAGS='-L /opt/ibm/sna/lib64' cargo build --release             *
//                                                                       *
// - on Windows                                                          *
//   set LIB=C:\ibmcs\w64cli\sdk64                                       *
//   cargo build --release                                               *
//                                                                       *
//  The rustaping and rust_tp programs are paired programs that send     *
//  pings via CPI-C API calls. This program will service both aping and  *
//  rustaping requests. The rust_tp transaction program will get invoked *
//  by the Communications Server when the aping or rustaping issues an   *
//  Allocate to the RUST_TP. The Communications Server uses the TP Load  *
//  Info to invoke the rust_tp executable.                               *
//                                                                       *
//  The statistics for the APING or RUSTAPING data received is written   *
//  to the output file /tmp/rust_tp.txt.                                 *
//                                                                       *
// ******* TP Definition:                                                *
// The rust_tp uses the TP name "RUST_TP". This can be defined           *
// using the snatpinstall utility to define the TP characteristics.      *
// Example TP settings for snatpinstall input file (rust_tp.tps):        *
//                                                                       *
//       [RUST_TP]                                                       *
//       PATH=/usr/local/bin/rust_tp                                     *
//       ENV=APPCTPN=RUST_TP                                             *
//       USERID=bin                                                      *
//       GROUP=sna                                                       *
//       TIMEOUT=60                                                      *
//       TYPE=QUEUED                                                     *
//                                                                       *
// Issue this command: snatpinstall -a rust_tp.tps                       *
//                                                                       *
// The TP definition will be loaded into the configuration.              *
// The TP should be defined as "Queued" so that incoming APING sessions  *
// and RUSTAPING sessions will be queued. A timeout of 60 seconds allows *
// faster execution for multiple requests. Any message received within   *
// the TIMEOUT specified length of time will be passed to the TP without *
// having to activate the program. This saves on program activiation and *
// initialization processing. If the TP does not receive an APING        *
// message in the specified amount of time, the TP will end and it will  *
// be restarted on the next APING or RUSTAPING session.                  *
//                                                                       *
// Issue this command to verify the TP definition is configured:         *
//                                                                       *
//      snatpinstall -q    or  snaadmin query_tp_load_info               *
//                                                                       *
// In order the run the TP as a application with permission to run as a  *
// SNA application. The user of the TP should be in the "sna" group. To  *
// create the required permission, Issue the following:                  *
//                                                                       *
//  >/usr/sbin/useradd -g sna snauser                                    *
//                                                                       *
// Copy the TP application to the /usr/local/bin directory (or where it  *
// is appropriate to run the TP application from). Make sure that the TP *
// definition in Comm. Server points to this same directory.             *
//                                                                       *
//  >cp -p rust_tp /usr/local/bin                                        *
//                                                                       *
// Set the TP owner to this new snauser:                                 *
//                                                                       *
//  >chown snauser.sna /usr/local/bin/rust_tp                            *
//                                                                       *
// The TP file should have this file listing for permissions and owner:  *
//                                                                       *
// >ls -l /usr/local/bin/provMgr                                         *
//  -rwxr-xr-x 1 bin sna 329161 Dec 2 17:36 /usr/local/bin/rust_tp        *
//                                                                       *
//************************************************************************
//                                                                       *
// � COPYRIGHT IBM Corporation 1989-2022                                 *
//                                                                       *
// ***********************************************************************
 
mod cpic;                  // include the CPI-C defintions for the API

use std::os::raw::{c_int,c_char};
use std::ascii::escape_default;
use std::str;
use std::env;
use std::process;
use chrono::{DateTime, Local};
use std::fs::File;
use std::io::Write;

//
// Error mapping table, this maps error codes to text strings for output
//
const ERR_CODE:[u8; 55] = [0,1,2,3,5,6,7,7,8,8,9,9,
                          10,11,17,18,19,20,21,22,23,
                          24,25,26,27,28,30,31,32,33,
                          34,35,36,37,38,39,40,41,42,
                          43,44,45,46,47,48,49,100,
                         130,131,132,133,134,135,136,137];

const RC_MSG:[&'static str; 55] = ["CM_OK","CM_ALLOCATE_FAILURE_NO_RETRY",
                                   "CM_ALLOCATE_FAILURE_RETRY",
                                   "CM_CONVERSATION_TYPE_MISMATCH",
                                   "CM_PIP_NOT_SPECIFIED_CORRECTLY",
                                   "CM_SECURITY_NOT_VALID",
                                   "CM_SYNC_LVL_NOT_SUPPORTED_LU",
                                   "CM_SYNC_LEVEL_NOT_SUPPORTED_LU",
                                   "CM_SYNC_LVL_NOT_SUPPORTED_PGM",
                                   "CM_SYNC_LEVEL_NOT_SUPPORTED_PGM",
                                   "CM_TPN_NOT_RECOGNIZED",
                                   "CM_TPN_NOT_RECOGNISED",
                                   "CM_TP_NOT_AVAILABLE_NO_RETRY",
                                   "CM_TP_NOT_AVAILABLE_RETRY",
                                   "CM_DEALLOCATED_ABEND",
                                   "CM_DEALLOCATED_NORMAL",
                                   "CM_PARAMETER_ERROR",
                                   "CM_PRODUCT_SPECIFIC_ERROR",
                                   "CM_PROGRAM_ERROR_NO_TRUNC",        
                                   "CM_PROGRAM_ERROR_PURGING",         
                                   "CM_PROGRAM_ERROR_TRUNC",           
                                   "CM_PROGRAM_PARAMETER_CHECK",       
                                   "CM_PROGRAM_STATE_CHECK",       
                                   "CM_RESOURCE_FAILURE_NO_RETRY", 
                                   "CM_RESOURCE_FAILURE_RETRY",    
                                   "CM_UNSUCCESSFUL",
                                   "CM_DEALLOCATED_ABEND_SVC",
                                   "CM_DEALLOCATED_ABEND_TIMER",
                                   "CM_SVC_ERROR_NO_TRUNC",
                                   "CM_SVC_ERROR_PURGING",
                                   "CM_SVC_ERROR_TRUNC",
                                   "CM_OPERATION_INCOMPLETE",
                                   "CM_SYSTEM_EVENT",
                                   "CM_OPERATION_NOT_ACCEPTED",
                                   "CM_CONVERSATION_ENDING",
                                   "CM_SEND_RCV_MODE_NOT_SUPPORTED",
                                   "CM_BUFFER_TOO_SMALL",
                                   "CM_EXP_DATA_NOT_SUPPORTED",
                                   "CM_DEALLOC_CONFIRM_REJECT",
                                   "CM_ALLOCATION_ERROR",
                                   "CM_RETRY_LIMIT_EXCEEDED",
                                   "CM_NO_SECONDARY_INFORMATION",
                                   "CM_SECURITY_NOT_SUPPORTED",
                                   "CM_SECURITY_MUTUAL_FAILED",
                                   "CM_CALL_NOT_SUPPORTED",
                                   "CM_PARM_VALUE_NOT_SUPPORTED",
                                   "CM_TAKE_BACKOUT",
                                   "CM_DEALLOCATED_ABEND_BO",
                                   "CM_DEALLOCATED_ABEND_SVC_BO",
                                   "CM_DEALLOCATED_ABEND_TIMER_BO",
                                   "CM_RESOURCE_FAIL_NO_RETRY_BO",
                                   "CM_RESOURCE_FAILURE_RETRY_BO",
                                   "CM_DEALLOCATED_NORMAL_BO",
                                   "CM_CONV_DEALLOC_AFTER_SYNCPT",
                                   "CM_INCLUDE_PARTNER_REJECT_BO"];

//  ########################################################
//
//  check_return( calling point string, file_ptr, error code) ->  return_code
//
//  Routine to check return code from CPI-C call and print errors if not CM_OK
//  When an error is reported, this routine will exit the program.
//
//  ########################################################
fn check_return(calling_point: &'static str, writefile: &mut File, return_code: c_int) -> c_int {
    if return_code != cpic::CM_OK {
   // iterate of error codes and get index into the matching return code's message
      for i in 0..55 {
          if return_code == ERR_CODE[i].into() {
       //      println!("error at calling point {}: return code: {}, value: {}",
            writeln!(writefile, "error at calling point {}: return code: {}, value: {}",
               calling_point,return_code,RC_MSG[i] ).unwrap();
            std::process::exit(1);
          }
      }      
 
    //  println!("error at calling point {}: return code {} ",
      writeln!(writefile, "error at calling point {}: return code {} ",
                calling_point,
                return_code).unwrap();
      std::process::exit(1);
      
    }

    return_code
}
//  ########################################################
//
//  receive_send_pings (conversation ID, file ptr, number of bytes sent/received) -> return_code    
//                   
//  Function to receive and then send frames back to partner
//
//  ########################################################
fn receive_send_pings(convid_ptr: *const c_char, writefile: &mut File, ptr_bytes: &mut c_int) -> c_int {

  let mut return_code: i32 = cpic::CM_OK;
  let mut recv_count: i32 = 0;
  let mut data_sent: i32  = 0;

  // Allocate 32K bytes
  let mut buffer = Vec::<c_char>::with_capacity(32763);

  let rcv_ptr = buffer.as_mut_ptr() as *mut c_char;


  let rcv_len = 32763;  // 32K

  let mut data_rcv_type: c_int = 0;

  let mut data_rcv_length: c_int = 0;

  let mut status_rcv: c_int = 0;

  let mut rcv_req_received_status: c_int  = 0;

  while status_rcv != cpic::CM_SEND_RECEIVED && return_code ==  cpic::CM_OK  {
     recv_count += 1;
     //
     // CMRCV to receive echo'd data from partner
     //
     unsafe {  // CMRCV call
        cpic::cmrcv(convid_ptr, rcv_ptr, &rcv_len, &mut data_rcv_type, 
                    &mut data_rcv_length, &mut status_rcv, 
                    &mut rcv_req_received_status, &mut return_code);
     }
     data_sent += data_rcv_length;
  }
 
//
//  If the CMRCV() does not indicate the partner deallocated the session
//  Then issue a CMSEND() to echo back the data received
//

  if return_code != cpic::CM_DEALLOCATED_NORMAL {

     if check_return("CMRCV rcv_snd_pings: ",writefile, return_code) ==  cpic::CM_OK {

  // CMSEND to send buffer initially indicating program level to partner

        let send_buff_ptr = buffer.as_mut_ptr() as *mut c_char;

        let send_buff_length = data_rcv_length;

        let mut send_req_received_status: i32 = 0;
        let mut send_count: i32 =0;

        while send_count < recv_count && 
           check_return("CMSEND call: ", writefile, return_code) ==  cpic::CM_OK {
           //
           // CMSEND to send buffer of bytes "lenght" to  partner 
           //
           unsafe {   // CMSEND Call 
              cpic::cmsend(convid_ptr, send_buff_ptr, &send_buff_length, 
                       &mut send_req_received_status, &mut return_code);
           }
           send_count +=1;
           data_sent += send_buff_length;
        }

        if check_return("CMSEND rcv_snd_pings: ",writefile, return_code) ==  cpic::CM_OK {

           *ptr_bytes += data_sent;

        }
     }
  }  
  return_code 
}
//  ########################################################
//
//  printbytes( address of vector of bytes, length) -> String
//
//  Routine to print bytes in array from slice of string 
//
//  ########################################################
fn printbyte(bs: &[u8], len: c_int) -> String {
    let mut visible = String::new();
    let len_usize:usize = len as usize; // convert i32 to usize
    for &b in bs {
        if visible.len() < len_usize {
           let part: Vec<u8> = escape_default(b).collect();
          visible.push_str(str::from_utf8(&part).unwrap());
        }  else {
          return visible;
        }
     }
    visible
}
//  ########################################################
//
//  make_slice( pointer to slice, length) -> vector
//
//   unsafe routine to make slice out of the receive buffer
//
//  ########################################################
unsafe fn make_slice<'a>(ptr: *const i8, len: usize) -> &'a [u8] {
    // place pointer address and length in contiguous memory
    let x: [usize; 2] = [ptr as usize, len];

    // cast pointer to array as pointer to slice
    let slice_ptr = &x as * const _ as *const &[u8];

    // dereference pointer to slice, so we get a slice
    *slice_ptr
}
//
// Global variables
//
static TP_NAME:        &str = "RUST_TP";
const  TP_NAME_LENGTH: c_int = 7;


//  ########################################################
//  -------------------------------------------------------------------------------
//
// main function to handle CMSLTP to set Local TP, CMACCP to accept incoming Allocates, 
//    Confirmation Startup and CMDEAL
//
//  -------------------------------------------------------------------------------
//  ########################################################
fn main() {

  // Define conversation ID variable
  let mut convid = Vec::<i8>::with_capacity(8); 
  let convid_ptr = convid.as_mut_ptr();

  let mut conversation_valid: bool = true;
  
  let mut return_code: c_int = 0;
  let tp_name_ptr = TP_NAME.as_ptr() as *const c_char;

  let mut write_file = File::create("/tmp/rust_tp.txt").unwrap();  // Log file for TP

  writeln!(&mut write_file, "IBM aping for RUST version 1.0: APPC echo test with timings.").unwrap();
 
  // CMSLTP call to specify local TP name (cmsltp) = RUSTP

  unsafe {
     cpic::cmsltp(tp_name_ptr, &TP_NAME_LENGTH, &mut return_code);
  }
  
  if check_return("CMSLTP call: ",&mut write_file, return_code) ==  cpic::CM_OK {

     while conversation_valid { 
  // Define the statistical results all in FP
         let mut total_bytes: c_int = 0;
         //
         // CMACCP to accept conversation request from partner on TP RUST_TP
         //
         unsafe {
            cpic::cmaccp(convid_ptr, &mut return_code);
         }

        if check_return("CMACCP call: ",&mut write_file, return_code) ==  cpic::CM_OK {
           let now: DateTime<Local> = Local::now();
           writeln!(&mut write_file, "").unwrap();  // Blank line before timestamp of start of session
           writeln!(&mut write_file, "Local time is: {}", now).unwrap();
        }
     //
     // Extract the local LU and partner LU for session and log them:
     //
        let mut lu_alias = String::from("         ");   // Initialize as blanks
        let lu_alias_ptr = lu_alias.as_ptr() as *mut c_char;
        //
        // CMELLN to extract the local LU and partner LU for session and log them:
        //
         unsafe {
              cpic::cmelln(convid_ptr, lu_alias_ptr, &mut return_code);
        }

        if check_return("CMELLN call: ",&mut write_file,return_code) !=  cpic::CM_OK {
            writeln!(&mut write_file, "Error getting Local LU Alias").unwrap();
        } 
        
     //
     // Extract the partner LU 
     //
        let mut partner_lu = String::from("                 ");   // Initialize as blanks
        let partner_lu_ptr = partner_lu.as_ptr() as *mut c_char;
        let mut partner_lu_length: c_int = 0;
        //
        // CMEPLN to extract the partner LU 
        //
        unsafe {
            cpic::cmepln(convid_ptr, partner_lu_ptr, &mut partner_lu_length, &mut return_code);
        }
        if check_return("CMEPLN call: ",&mut write_file,return_code) !=  cpic::CM_OK {
           writeln!(&mut write_file, "Error getting Partner LU ").unwrap();
        } 
        writeln!(&mut write_file, "  Local LU ALias: {} connected to Partner LU: {}", 
                 lu_alias, partner_lu).unwrap();

     //   
     // Define receive call #1, partner level of APING
     // 
        let mut rcv_buff = Vec::<c_char>::with_capacity(100);
        rcv_buff.resize(100,0);   // Fills and sets vector to 100 byte length

        let rcv_ptr = rcv_buff.as_mut_ptr();
  
        let rcv_len = 100;

        let mut data_rcv_type: c_int = 0;

        let mut data_rcv_length: c_int = 0;

        let mut status_rcv: c_int = 0;

        let mut rcv_req_received_status: c_int  = 0;
        //
        // CMRCV to receive initial exchange to indicate program level of partner
        //
        unsafe {
           cpic::cmrcv(convid_ptr, rcv_ptr, &rcv_len, &mut data_rcv_type, &mut data_rcv_length,
                &mut status_rcv, &mut rcv_req_received_status, &mut return_code);
         }

         if check_return("CMRCV call #1: ",&mut write_file,return_code) ==  cpic::CM_OK {
            if data_rcv_type == cpic::CM_COMPLETE_DATA_RECEIVED {

  // CMSEND to send buffer initially indicating program level to partner

       const FIRSTREC_DATA: &[u8] = &[
         0x01, 0x02, 0x2c, 0x02, // i32
         0x52, 0x55, 0x53, 0x54, // RUST
         0x5f, 0x54, 0x50, 0x20, // _TP
         0x66, 0x6f, 0x72, 0x20, // for
         0x4c, 0x69, 0x6e, 0x75, // Linu
         0x78, 0x20, 0x31, 0x2e, // x 1. 
         0x30, 0x20, 0x20, 0x20, // 0 
       ];
    // Set C string for Symbolic name
       let send_buff_ptr = FIRSTREC_DATA.as_ptr() as *const i8;

       let send_buff_length = 28;

       let mut send_req_received_status: i32 = 0;
            //
            // CMSEND to send initial exchange to indicate program local level of code
            //
       unsafe {
          cpic::cmsend(convid_ptr, send_buff_ptr, &send_buff_length, 
                       &mut send_req_received_status, &mut return_code);
       }

       if check_return("CMSEND call #1: ",&mut write_file,return_code) ==  cpic::CM_OK {
     // print the received level from partner doing APING or RUSTAPING
           let print_length: c_int = data_rcv_length + 9;  // add 9 bytes because first 4 bytes have 3 white space
                                                            // bytes that get displayed as HEX \x0y, where y is digit. 
           writeln!(&mut write_file, "  Program startup and Confirm complete").unwrap();
           unsafe {
                writeln!(&mut write_file,"  {}",
                     printbyte(make_slice(rcv_ptr, print_length as usize),print_length)).unwrap(); }
           } else {
             writeln!(&mut write_file, " ** Program startup and Confirm failure: Data not completely received!!").unwrap();
              std::process::exit(1);           
           }
         }
       }
       return_code = cpic::CM_OK;

       while return_code == cpic::CM_OK {

  //  Receive until return code is not CM_OK. Check that it is CM_DEALLOCATE_NORMAL, otherwise
  //   report it as error
  //
           return_code = receive_send_pings(convid_ptr, &mut write_file, &mut total_bytes);  

           if return_code != cpic::CM_OK  &&  return_code != cpic::CM_DEALLOCATED_NORMAL {
               writeln!(&mut write_file," Error in receiving and sending pings, {} ", return_code).unwrap();
               conversation_valid = false;
           }
       } 
       writeln!(&mut write_file," Total bytes received from partner: {} ", total_bytes).unwrap();
       writeln!(&mut write_file, "").unwrap();

     }  // end while convsation is valid
  } // If Specify TP Name

} // end of main function

