// ***********************************************************************
//  Module:    rustaping                                                 *
//                                                                       *
//  Purpose:   Example program sending APPC pings using CPI-C API        *
//                                                                       *
// To compile this program use the following cargo build commands:       *
//                                                                       *
// - on Linux                                                            *
//   export LD_LIBRARY_PATH=/opt/ibm/sna/lib64>/usr/lib64                *
//   export LD_RUN_PATH=/opt/ibm/sna/lib64/usr/lib64                     *
//   RUSTFLAGS='-L /opt/ibm/sna/lib64' cargo build --release             *
//                                                                       *
// - on Windows                                                          *
//   set LIB=C:\ibmcs\w64cli\sdk64                                       *
//   cargo build --release                                               *
//                                                                       *
//   This file, lib.rs, contains the input parameters and global         *
//   parameters for the rustaping program.                               *
//                                                                       *
// ***********************************************************************
//                                                                       *
// ? COPYRIGHT IBM Corporation 1989-2022                                 *
//                                                                       *
// ***********************************************************************
 
#![allow(non_upper_case_globals)]
#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(unused_imports)]

use std::error::Error;
use std::env;

// Default settings for input variables
const  MAX_PINGS:       i32 = 3;
const  DEFAULT_SIZE:    i32 = 1000;
const  REPEAT_COUNT:    i32 = 1;
static TP_NAME:        &str = "APINGD  ";
const  SHOW_LINE_NUMS: bool = false;     



//
//  Config structure holds the default/input parameters from startup
//
pub struct Config {
    pub  buffsize: i32,
    pub  iteration: i32,
    pub  repeat_count: i32,
    pub  symbolic_name: String,
    pub  line_numbers: bool,
}

//
// new implementation will create new Config and return defaults/input parameters or Err
//
impl Config {
    pub fn new(args: &[String]) -> Result<Config, &'static str> {
      let mut buffsize: i32 = DEFAULT_SIZE;
      let mut iteration: i32 = MAX_PINGS;
      let mut repeat_count: i32 = REPEAT_COUNT;
      let mut symbolic_name: String = TP_NAME.to_string();
      let mut line_numbers: bool = SHOW_LINE_NUMS;        
      
      let mut index: usize = 1; 
      while index < args.len() {
          match &args[index]as &str {  // size parameter
            "-s" => { buffsize = match args[index+1].parse() {
                         Ok(n) => {
                            n
                         },
                         Err(_) => {
                           eprintln!("error: argument {} not an integer",index+1);
                           help();
                           return Err("Invalid argument entered for size");
                         },
                      };
                      index += 2;
                   },
                 // iteration parameter
           "-c" => { repeat_count = match args[index+1].parse() {
                         Ok(n) => {
                            n
                         },
                         Err(_) => {
                           eprintln!("error: argument {} not an integer",index+1);
                           help();
                           return Err("Invalid argument entered for repeat count");
                         },
                      };
                      index += 2;
                   },
                 // iteration parameter
            "-i" => { iteration = match args[index+1].parse() {
                         Ok(n) => {
                            n
                         },
                         Err(_) => {
                            eprintln!("error: argument {} not an integer", index+1);
                            help();
                            return Err("Invalid argument entered for iteration");
                         },
                      };
                      index += 2;
                   },
                 // input flag to indicate line numbers needed on output
            "-#" => { line_numbers = true; 
                      index += 1; },
                  // CPI-C Side Info name
            _ => {  symbolic_name = format!("{:width$}", args[index], width=8); // This pads right blanks to string
                    if !symbolic_name.chars().next().unwrap().is_ascii_alphabetic() {  
                      
                       println!("\n error: destination argument {} not valid character\n", index);
                       help();
                       return Err("Invalid argument entered for CPI-C-Side-Info-Name");
                    }
                    index = args.len(); 
                  },
           };
      }
      Ok(Config { buffsize, iteration, repeat_count, symbolic_name, line_numbers })
    }
}

fn help() {
    println!("usage:
rustaping [-s size]  [-i iterations] [-# (print line numbers on output)] [CPI-C-Side-Info-Name]
defaults: size is 1000, iterations is 3, no line numbers, CPI-C-Side-Info-Name is APINGD    
");
}

