
// ***********************************************************************
//  Module:    rustaping                                                 *
//                                                                       *
//  Purpose:   Example program sending APPC pings using CPI-C API        *
//                                                                       *
// To compile this program use the following cargo build commands:       *
//                                                                       *
// - on Linux                                                            *
//   export LD_LIBRARY_PATH=/opt/ibm/sna/lib64:/usr/lib64                *
//   export LD_RUN_PATH=/opt/ibm/sna/lib64:/usr/lib64                    *
//   RUSTFLAGS='-L /opt/ibm/sna/lib64' cargo build --release             *
//                                                                       *
// - on Windows                                                          *
//   set LIB=C:\ibmcs\w64cli\sdk64                                       *
//   cargo build --release                                               *
//                                                                       *
//  The rustaping and rust_tp programs are paired programs that send     *
//  pings via CPI-C API calls. The rustaping parameters are:             *
//       -i iterations - Number of times to send pings, default is 3     *
//       -c count      - Number of times each iteration to send frames   *
//                       default is 1                                    *
//       -s size       - size in bytes of each frame in a ping           *
//                       default is 1000 bytes                           *
//       -#            - display line numbers on output, default is off  *
//       CPI-C symbolic name - required                                  *
//                                                                       *
//  The rustaping uses the CPIC_Side_Info record in the Communications   *
//  Server to list the partner LU, mode, local LU alias and security.    *
//  The program consists of 4 subroutines and the main function.         *
//  This example program shows how a Transaction Program (TP) initiates  *
//  a TP, allocates a session and sends/receives data. Finally, it       *
//  deallocates the session once the number of pings has completed.      *
//  The output reports the number of bytes and the bandwidth rate on     *
//  each ping. At the end of the program, it prints the total rate       *
//  realized.                                                            *
//                                                                       *
//  The CPIC_SIDE_INFO record defined will have a minimum of the         *
//  following parameters defined.                                        *
//       - symbolic destination name or label for the CPIC_SIDE_INFO     *
//       - Partner LU name in fully qualified LU name format             *
//       - Transaction Program name:                                     *
//             RUST_TP to invoke the rust_tp program OR                  *
//             APINGD  to invoke the APINGD TP standard in SNA nodes     *
//                                                                       *
//  Example CPIC_SIDE_INFO record:                                       *
//     sym_dest_name = RUST_TPM                                          *
//     description = "Rust example TP"                                   *
//     partner_lu_name = USIBMNM.MTRSLNN                                 *
//     tp_name_type = APPLICATION_TP                                     *
//     tp_name = RUST_TP                                                 *
//     mode_name = #BATCH                                                *
//     conversation_security_type = NONE                                 *
//     security_user_id = ""                                             *
//     security_password = ""                                            *
//     lu_alias = ""                                                     *
//                                                                       *
//  Example invocation of rustaping:                                     *
// ./rustaping -i 5 -s 100 -c 5 -# RUST_TPM                              *
// IBM aping for RUST version 1.0: APPC echo test with timings.          *
// Allocate duration:   164.6 ms                                         *
//   Local LU ALias: cstestsl connected to Partner LU: USIBMNM.MTRSLNN   *
// Program startup and Confirm duration:   116.7 ms                      *
//                                                                       *
//  Connected to partner running on: RUST_TP for Linux 1.0               *
// #1 Duration(ms): 402.1  Data Sent(bytes): 1000  Data Rate(KB/s) 2.4  (Mb/s) 0.019 
// #2 Duration(ms): 396.5  Data Sent(bytes): 1000  Data Rate(KB/s) 2.5  (Mb/s) 0.020 
// #3 Duration(ms): 383.8  Data Sent(bytes): 1000  Data Rate(KB/s) 2.5  (Mb/s) 0.020 
// #4 Duration(ms): 385.2  Data Sent(bytes): 1000  Data Rate(KB/s) 2.5  (Mb/s) 0.020 
// #5 Duration(ms): 380.0  Data Sent(bytes): 1000  Data Rate(KB/s) 2.6  (Mb/s) 0.021 
// -- Totals:                                                            *
// Duration(ms): 1947.6  Total Data (bytes): 5000  Data Rate(KB/s) 2.5  (Mb/s) 0.020 
//                                                                       *
// *********************************************************************** 
//                                                                       *
// ? COPYRIGHT IBM Corporation 1989-2022                                 *
//                                                                       *
// ***********************************************************************
 
mod cpic;          // include the CPI-C defintions for the API

use std::os::raw::{c_int,c_char};
use std::ascii::escape_default;
use std::str;
use std::time::{Instant};
use std::env;
use std::process;

//
// Error mapping table, this maps error codes to text strings for output
//
const ERR_CODE:[u8; 55] = [0,1,2,3,5,6,7,7,8,8,9,9,
                          10,11,17,18,19,20,21,22,23,
                          24,25,26,27,28,30,31,32,33,
                          34,35,36,37,38,39,40,41,42,
                          43,44,45,46,47,48,49,100,
                         130,131,132,133,134,135,136,137];

const RC_MSG:[&'static str; 55] = ["CM_OK","CM_ALLOCATE_FAILURE_NO_RETRY",
                                   "CM_ALLOCATE_FAILURE_RETRY",
                                   "CM_CONVERSATION_TYPE_MISMATCH",
                                   "CM_PIP_NOT_SPECIFIED_CORRECTLY",
                                   "CM_SECURITY_NOT_VALID",
                                   "CM_SYNC_LVL_NOT_SUPPORTED_LU",
                                   "CM_SYNC_LEVEL_NOT_SUPPORTED_LU",
                                   "CM_SYNC_LVL_NOT_SUPPORTED_PGM",
                                   "CM_SYNC_LEVEL_NOT_SUPPORTED_PGM",
                                   "CM_TPN_NOT_RECOGNIZED",
                                   "CM_TPN_NOT_RECOGNISED",
                                   "CM_TP_NOT_AVAILABLE_NO_RETRY",
                                   "CM_TP_NOT_AVAILABLE_RETRY",
                                   "CM_DEALLOCATED_ABEND",
                                   "CM_DEALLOCATED_NORMAL",
                                   "CM_PARAMETER_ERROR",
                                   "CM_PRODUCT_SPECIFIC_ERROR",
                                   "CM_PROGRAM_ERROR_NO_TRUNC",        
                                   "CM_PROGRAM_ERROR_PURGING",         
                                   "CM_PROGRAM_ERROR_TRUNC",           
                                   "CM_PROGRAM_PARAMETER_CHECK",       
                                   "CM_PROGRAM_STATE_CHECK",       
                                   "CM_RESOURCE_FAILURE_NO_RETRY", 
                                   "CM_RESOURCE_FAILURE_RETRY",    
                                   "CM_UNSUCCESSFUL",
                                   "CM_DEALLOCATED_ABEND_SVC",
                                   "CM_DEALLOCATED_ABEND_TIMER",
                                   "CM_SVC_ERROR_NO_TRUNC",
                                   "CM_SVC_ERROR_PURGING",
                                   "CM_SVC_ERROR_TRUNC",
                                   "CM_OPERATION_INCOMPLETE",
                                   "CM_SYSTEM_EVENT",
                                   "CM_OPERATION_NOT_ACCEPTED",
                                   "CM_CONVERSATION_ENDING",
                                   "CM_SEND_RCV_MODE_NOT_SUPPORTED",
                                   "CM_BUFFER_TOO_SMALL",
                                   "CM_EXP_DATA_NOT_SUPPORTED",
                                   "CM_DEALLOC_CONFIRM_REJECT",
                                   "CM_ALLOCATION_ERROR",
                                   "CM_RETRY_LIMIT_EXCEEDED",
                                   "CM_NO_SECONDARY_INFORMATION",
                                   "CM_SECURITY_NOT_SUPPORTED",
                                   "CM_SECURITY_MUTUAL_FAILED",
                                   "CM_CALL_NOT_SUPPORTED",
                                   "CM_PARM_VALUE_NOT_SUPPORTED",
                                   "CM_TAKE_BACKOUT",
                                   "CM_DEALLOCATED_ABEND_BO",
                                   "CM_DEALLOCATED_ABEND_SVC_BO",
                                   "CM_DEALLOCATED_ABEND_TIMER_BO",
                                   "CM_RESOURCE_FAIL_NO_RETRY_BO",
                                   "CM_RESOURCE_FAILURE_RETRY_BO",
                                   "CM_DEALLOCATED_NORMAL_BO",
                                   "CM_CONV_DEALLOC_AFTER_SYNCPT",
                                   "CM_INCLUDE_PARTNER_REJECT_BO"];

//  ########################################################
//
//  check_return( calling point string, error code) ->  return_code
//
//  Routine to check return code from CPI-C call and print errors if not CM_OK
//  When an error is reported, this routine will exit the program.
//
//  ########################################################
fn check_return(calling_point: &'static str, return_code: c_int) -> c_int {
    if return_code != cpic::CM_OK {
   // iterate of error codes and get index into the matching return code's message
      for i in 0..55 {
          if return_code == ERR_CODE[i].into() {
             println!("error at calling point {}: return code: {}, value: {}",
               calling_point,return_code,RC_MSG[i] );
            std::process::exit(1);
          }
      }      
 
      println!("error at calling point {}: return code {} ",
                calling_point,
                return_code);
      std::process::exit(1);
      
    }

    return_code
}
//  ########################################################
//
//  send_receive_pings (conversation ID, number of bytes to send, number of times to 
//                      repeat sending, duration in microseconds, pointer to bytes 
//                      to send, count is the number to display on output, flag to 
//                      display output number)   
//                   
//  Function to send and receive frames to partner
//
//  ########################################################
fn send_receive_pings(convid_ptr: *const c_char, num_of_bytes:i32, repeat_count:i32,
                      ptr_duration: &mut f64, ptr_bytes: &mut f64,
                      count: i32, display_line_num: bool) {

  let mut return_code: i32 = 0;
  let mut send_count: i32 = 0;    /* repeat send or receive count */
  let mut data_sent: f64  = 0.0;

  // Allocate 32K bytes
  let mut buffer = vec![0_u8; 32763];
  let send_buff_ptr = buffer.as_ptr() as *const c_char;

  let send_buff_length = num_of_bytes;

  let mut send_req_received_status: i32 = 0;
   
  let timestart = Instant::now();

  while send_count < repeat_count && check_return("CMSEND call: ",return_code) ==  cpic::CM_OK {
     send_count += 1;
     //
     // CMSEND to send buffer of bytes "length" to  partner 
     //
     unsafe {   // CMSEND Call  
        cpic::cmsend(convid_ptr, send_buff_ptr, &send_buff_length, 
                   &mut send_req_received_status, &mut return_code);
      }
     data_sent += f64::from(send_buff_length);
  }

  if check_return("CMSEND call: ",return_code) ==  cpic::CM_OK {
    
  // Define receive call #1, partner level of APING
  
     let rcv_ptr = buffer.as_mut_ptr() as *mut c_char;

     let rcv_len = 32763;  // 32K

     let mut data_rcv_type: c_int = 0;

     let mut data_rcv_length: c_int = 0;

     let mut status_rcv: c_int = 0;

     let mut rcv_req_received_status: c_int  = 0;

     send_count = 0;
     while send_count < repeat_count && check_return("CMRCV call: ",return_code) ==  cpic::CM_OK {
       send_count += 1;
        //
        // CMRCV to receive echo'd data from partner
        //
        unsafe {  // CMRCV call
        cpic::cmrcv(convid_ptr, rcv_ptr, &rcv_len, &mut data_rcv_type, &mut data_rcv_length,
                 &mut status_rcv, &mut rcv_req_received_status, &mut return_code);
        }
        data_sent += f64::from(data_rcv_length);
        
     }

     if check_return("CMRCV call #1: ",return_code) ==  cpic::CM_OK {
        let elapsed_time: f64 = timestart.elapsed().as_secs_f64() * 1000.0; //to get milliseconds
        let kbs_rate: f64     = (( data_sent / 1024.0 ) / elapsed_time) * 1000.0;
        let mbs_rate: f64     = kbs_rate * 8.0 / 1000.0;

         // Format as milliseconds rounded down
        if display_line_num {
            println!("#{} Duration(ms): {:.1}  Data Sent(bytes): {:.0}  Data Rate(KB/s) {:.1}  (Mb/s) {:.3} ", 
                 count, elapsed_time, data_sent, kbs_rate, mbs_rate );        
        }else {
            println!("Duration(ms): {:.1}  Data Sent(bytes): {:.0}  Data Rate(KB/s) {:.1}  (Mb/s) {:.3} ", 
                 elapsed_time, data_sent, kbs_rate, mbs_rate );
        }
          // Add current duration and bytes to total for tracking full transaction
        *ptr_duration += elapsed_time;
        *ptr_bytes += data_sent;

     }
  }   
}
//  ########################################################
//
//  printbytes( address of vector of bytes, length) -> String
//
//  Routine to print bytes in array from slice of string 
//
//  ########################################################
fn printbyte(bs: &[u8], len: c_int) -> String {
    let mut visible = String::new();
    let len_usize:usize = len as usize;  // convert i32 to usize 
    let mut b_count = 0;
    for &b in bs {
        if b_count > 3 {  // skip first 4 bytes
           if visible.len() < len_usize {
             let part: Vec<u8> = escape_default(b).collect();
             visible.push_str(str::from_utf8(&part).unwrap());
           }  else {
             return visible;
           }
        }
        b_count +=1;
    }
    visible
}
//  ########################################################
//
//  make_slice( pointer to slice, length) -> vector
//
//   unsafe routine to make slice out of the receive buffer
//
//  ########################################################
unsafe fn make_slice<'a>(ptr: *const c_char, len: usize) -> &'a [u8] {
    // place pointer address and length in contiguous memory
    let x: [usize; 2] = [ptr as usize, len];

    // cast pointer to array as pointer to slice
    let slice_ptr = &x as * const _ as *const &[u8];

    // dereference pointer to slice, so we get a slice
    *slice_ptr
}

//  ########################################################
//  -------------------------------------------------------------------------------
//
// main function to handle CMINIT, CMALLOC, Confirmation Startup and CMDEAL
//
//  -------------------------------------------------------------------------------
//  ########################################################
fn main() {

  // Define the statistical results all in FP
  let mut total_duration:f64 = 0.0;
  let mut total_bytes: f64 = 0.0;
  let total_kbs: f64;
  let total_mbs:f64;
//
// config is a structure, see lib.rs, that has defaults or passed in arguments for 
// TP_NAME, size, iterations and showing line numbers
//
  let args: Vec<String> = env::args().collect();
  let config = rustaping::Config::new(&args).unwrap_or_else(|err| {
       eprintln!("Problem parsing arguments: {}", err);
       process::exit(1);
  });

  // Define conversation ID variable
  let mut convid = Vec::<c_char>::with_capacity(8); 
  let convid_ptr = convid.as_mut_ptr();
  
  let mut return_code: i32 = 0;
  let sym_name_ptr = config.symbolic_name.as_ptr() as *const c_char;

  println!("IBM aping for RUST version 1.0: APPC echo test with timings.");
  
 
  // CMINIT call to start conversation

  unsafe {
     cpic::cminit(convid_ptr, sym_name_ptr, &mut return_code);
  }
  
  if check_return("CMINIT call: ",return_code) ==  cpic::CM_OK {
     let timestart = Instant::now();

  // CMALLC to allocate conversation to partner

     unsafe {
         cpic::cmallc(convid_ptr, &mut return_code);
     }

     if check_return("CMALLC call: ",return_code) ==  cpic::CM_OK {
        let elapsed_time: f64 = timestart.elapsed().as_secs_f64() * 1000.0; //to get milliseconds
        println!("Allocate duration:   {:.1} ms",elapsed_time);
     }
  }

  //
  // CMELLN to extract the local LU and partner LU for session and log them:
  //
  let lu_alias = String::from("         ");   // Initialize as blanks
  let lu_alias_ptr = lu_alias.as_ptr() as *mut c_char;

  unsafe {
        cpic::cmelln(convid_ptr, lu_alias_ptr, &mut return_code);
  }

  if check_return("CMELLN call: ", return_code) !=  cpic::CM_OK {
      println!("Error getting Local LU Alias");
  } 
        
  //
  // CMEPLN to extract the partner LU 
  //
  let partner_lu = String::from("                 ");   // Initialize as blanks
  let partner_lu_ptr = partner_lu.as_ptr() as *mut c_char;
  let mut partner_lu_length: c_int = 0;

  unsafe {
      cpic::cmepln(convid_ptr, partner_lu_ptr, &mut partner_lu_length, &mut return_code);
  }
  if check_return("CMEPLN call: ", return_code) !=  cpic::CM_OK {
       println!("Error getting Partner LU ");
  } 
  println!("  Local LU ALias: {} connected to Partner LU: {}",lu_alias, partner_lu);

  //
  // CMSEND to send buffer initially indicating program level to partner
  //
  const FIRSTREC_DATA: &[u8] = &[
    0x01, 0x02, 0x2c, 0x02, // i32
    0x52, 0x55, 0x53, 0x54, // RUST
    0x41, 0x50, 0x49, 0x4e, // APIN
    0x47, 0x20, 0x66, 0x6f, // G fo
    0x72, 0x20, 0x4c, 0x69, // f Li
    0x6e, 0x75, 0x78, 0x20, // nux 
    0x31, 0x2e, 0x30, 0x20, // 1.0 
  ];
  // Set C string for Symbolic name
  let send_buff_ptr = FIRSTREC_DATA.as_ptr() as *const c_char;

  let send_buff_length = 28;

  let mut send_req_received_status: i32 = 0;
      
  let timestart = Instant::now(); 

  unsafe {
     cpic::cmsend(convid_ptr, send_buff_ptr, &send_buff_length, &mut send_req_received_status, &mut return_code);
  }

  if check_return("CMSEND call #1: ",return_code) ==  cpic::CM_OK {

     // Define receive call #1, partner level of APING
     let mut rcv_buff = Vec::<c_char>::with_capacity(100);
     let rcv_ptr = rcv_buff.as_mut_ptr();
  
     let rcv_len = 100;

     let mut data_rcv_type: c_int = 0;

     let mut data_rcv_length: c_int = 0;

     let mut status_rcv: c_int = 0;

     let mut rcv_req_received_status: c_int  = 0;
     //
     // CMRCV to receive initial exchange to indicate program level of partner
     //
     unsafe {
        cpic::cmrcv(convid_ptr, rcv_ptr, &rcv_len, &mut data_rcv_type, &mut data_rcv_length,
                 &mut status_rcv, &mut rcv_req_received_status, &mut return_code);
     }

     if check_return("CMRCV call #1: ",return_code) ==  cpic::CM_OK {
        if data_rcv_type == cpic::CM_COMPLETE_DATA_RECEIVED {
           let elapsed_time: f64 = timestart.elapsed().as_secs_f64() * 1000.0; //to get milliseconds
           println!("Program startup and Confirm duration:   {:.1} ms",elapsed_time);
           println!("");

           let partner_level_strlen = data_rcv_length-1;  // first 4 bytes are binary indicators
           unsafe { println!(" Connected to partner running on: {}",
               printbyte(make_slice(rcv_ptr, partner_level_strlen as usize),partner_level_strlen));
           }
        }
        else {
           println!(" ** Program startup and Confirm failure: Data not completely received!!");
           std::process::exit(1);           
        }
     }

  }    
  //
  //  Send specified number of pings, repeating each ping specified times and get running statistics
  //  This is the main loop for sending and receiving from partner in CPI-C Side Info record
  //
    let mut index = 0;

    while index < config.iteration {

        send_receive_pings(convid_ptr, config.buffsize, config.repeat_count, 
                           &mut total_duration, &mut total_bytes,
                           index+1, config.line_numbers);  // pass index starting at one
        index += 1;

    }    
  //
  // CMDEAL to deallocate the session at the end of the run
  //
  unsafe {  //CMDEAL call  
     cpic::cmdeal(convid_ptr,&mut return_code);
  }
  //
  // Display to total statistics 
  //
  if check_return("CMDEAL call: ",return_code) ==  cpic::CM_OK {
      println!("-- Totals: ");
      total_kbs  = (( total_bytes / 1024.0 ) / total_duration) * 1000.0;
      total_mbs  = total_kbs * 8.0 / 1000.0;
      println!("Duration(ms): {:.1}  Total Data (bytes): {:.0}  Data Rate(KB/s) {:.1}  (Mb/s) {:.3} ",
                 total_duration, total_bytes, total_kbs, total_mbs );
  }
  
}
